//
// xitalkbiff.cc
//


#include "socket-stream.h"
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <ctype.h>


static String default_resources[]=
{
  "*Title: XItalkBiff",
  NULL,
};


static XrmOptionDescRec options[]=
{
{"--help",   ".help",          XrmoptionNoArg,  (XPointer)"True"},
{"-help",    ".help",          XrmoptionNoArg,  (XPointer)"True"},
{"-h",       ".help",          XrmoptionNoArg,  (XPointer)"True"},
{"-l",       ".landscape",     XrmoptionNoArg,  (XPointer)"True"},
{"-nl",      ".landscape",     XrmoptionNoArg,  (XPointer)"False"},
{"-i",       ".italkPipe",     XrmoptionSepArg, (XPointer)NULL},
{"-fn8",     ".font",          XrmoptionSepArg, (XPointer)NULL},
{"-fnAscii", ".font",          XrmoptionSepArg, (XPointer)NULL},
{"-fn16",    ".kanjiFont",     XrmoptionSepArg, (XPointer)NULL},
{"-fk",      ".kanjiFont",     XrmoptionSepArg, (XPointer)NULL},
{"-fnKanji", ".kanjiFont",     XrmoptionSepArg, (XPointer)NULL},
{"-fr",      ".romanKanaFont", XrmoptionSepArg, (XPointer)NULL},
{"-fnKana",  ".romanKanaFont", XrmoptionSepArg, (XPointer)NULL},
{"-star",    ".star",          XrmoptionSepArg, (XPointer)NULL},
{"-bg2",     ".background2",   XrmoptionSepArg, (XPointer)NULL},
{"-p",       ".popup",         XrmoptionNoArg,  (XPointer)"True"},
{"-np",      ".popup",         XrmoptionNoArg,  (XPointer)"False"},
{"-pt",      ".popupTime",     XrmoptionSepArg, (XPointer)NULL},
};


static struct AppData
{
  String italkpipe;
  String fontAscii;
  String fontKanji;
  String fontKana;
  String foreground;
  String star;
  String background;
  String background2;
  Bool   help;
  Bool   landscape;
  Bool   popup;
  int    popup_time;
}app_data;


static XtResource resources[]=
{
{ "italkPipe",    "ItalkPipe",     XtRString,  sizeof(String),  XtOffset(AppData*, italkpipe),   XtRString,    (XtPointer)"ecc-as50.komaba.ecc.u-tokyo.ac.jp:12344"},
{ "font",         "Font",          XtRString,  sizeof(String),  XtOffset(AppData*, fontAscii),   XtRString,    (XtPointer)"-adobe-helvetica-bold-r-*-*-14-*-*-*-*-*-*-*"},
{ "romanKanaFont","RomanKanaFont", XtRString,  sizeof(String),  XtOffset(AppData*, fontKana),    XtRString,    (XtPointer)"-misc-*-jisx0201.1976-*"},
{ "kanjiFont",    "KanjiFont",     XtRString,  sizeof(String),  XtOffset(AppData*, fontKanji),   XtRString,    (XtPointer)"k14"},
{ "foreground",   "Foreground",    XtRString,  sizeof(String),  XtOffset(AppData*, foreground),  XtRString,    (XtPointer)"White"},
{ "star",         "Foreground",    XtRString,  sizeof(String),  XtOffset(AppData*, star),        XtRString,    (XtPointer)"Yellow"},
{ "background",   "Background",    XtRString,  sizeof(String),  XtOffset(AppData*, background),  XtRString,    (XtPointer)"gray50"},
{ "background2",  "Background",    XtRString,  sizeof(String),  XtOffset(AppData*, background2), XtRString,    (XtPointer)"gray40"},
{ "help",         "Help",          XtRBoolean, sizeof(Boolean), XtOffset(AppData*, help),        XtRImmediate, (XtPointer)False},
{ "landscape",    "Landscape",     XtRBoolean, sizeof(Boolean), XtOffset(AppData*, landscape),   XtRImmediate, (XtPointer)False},
{ "popup",        "Popup",         XtRBoolean, sizeof(Boolean), XtOffset(AppData*, popup),       XtRImmediate, (XtPointer)False},
{ "popupTime",    "PopupTime",     XtRInt,     sizeof(int),     XtOffset(AppData*, popup_time),  XtRImmediate, (XtPointer)10},
};


static Widget       shell;
static Window       w=0;
static GC           gc=0;
static Pixmap       pxm=0;
static GC           pxmgc=0;
static Display      *d=NULL;
static XtAppContext acr;
static XFontStruct  *fsAscii;
static XFontStruct  *fsKanji;
static XFontStruct  *fsKana;

//static Pixmap     icon, mask;
static XColor       fore, star, back, back2;
static Atom         atom_wm_protocols, atom_wm_delete_window;


#define MAX_WHO 30


class Who
{
public:
  char *who[MAX_WHO];
  inline Who(void)
  {
    for (int i=0; i<MAX_WHO; i++)
      who[i]=NULL;
  }
  inline ~Who(void)
  {
    Clear();
  }
  inline void Clear(void)
  {
    for (int i=0; i<MAX_WHO; i++)
      delete_string(who[i]);
  }
  void Add(char *_who)
  {
    int i;
    for (i=0; i<MAX_WHO; i++)
    {
      if (!who[i]) break;
      if (strcmp(who[i], _who)==0) return;
    }
    if (i==MAX_WHO) return;
    who[i]=new_string(_who);
  }
  inline int N(void)
  {
    int n=0;
    for (int i=0; i<MAX_WHO; i++)
      if (who[i]) n++;
    return n;
  }
}who;


static void on8(char *s, int len) // 8 bit ܤΩƤ
{
  while(len) *(s++)|=0x80, len--;
}


static void off8(char *s, int len) // 8 bit ܤ򥯥ꥢ
{
  while(len) *(s++)&=0x7f, len--;
}


static int GetFontHeight(void)
{
  return max(max(fsAscii->max_bounds.ascent+fsAscii->max_bounds.descent,
		 fsKana ->max_bounds.ascent+fsKana ->max_bounds.descent),
	     fsKanji->max_bounds.ascent+fsKanji->max_bounds.descent);
}


enum EUCtype { Ascii, Kanji, Kana, Star, Esc, Null };


EUCtype GetEUCtype(char *c)
{
  if (c[0]==0) return Null;
  else if (c[0]=='\x1b') return Esc;
  else if (c[0]>0) return Ascii;
  else if (c[0]=='\x8e') return Kana;
  else if ((c[0]=='\xa1' && c[1]=='\xfa') ||
	   (c[0]=='\xa1' && c[1]=='\xf9')) return Star;
  return Kanji;
}


static int GetWidth(char *str)
{
  int width=0;
  char *c=str;
  char *s=str;
  EUCtype et=GetEUCtype(str);
  
  int dummy;
  XCharStruct cs;
  
  while (1)
  {
    if (et==Ascii)
    {
      if (Ascii!=GetEUCtype(c))
      {
	XTextExtents(fsAscii, s, c-s, &dummy, &dummy, &dummy, &cs);
	width+=cs.width;
	et=GetEUCtype(c);
	s=c;
      }
      else c++;
    }
    else if (et==Esc)
    {
      if (*c=='m')
      {
	c++;
	et=GetEUCtype(c);
	s=c;
      }
      else if (*c=='\0') et=Null;
      else c++;
    }
    else if (et==Kanji)
    {
      if (Kanji!=GetEUCtype(c))
      {
	off8(s, c-s);
	XTextExtents16(fsKanji, (XChar2b*)s, (c-s)/2, &dummy, &dummy, &dummy, &cs);
	on8(s, c-s);
	width+=cs.width;
	et=GetEUCtype(c);
	s=c;
      }
      else c+=2;
    }
    else if (et==Star)
    {
      if (Star!=GetEUCtype(c))
      {
	off8(s, c-s);
	XTextExtents16(fsKanji, (XChar2b*)s, (c-s)/2, &dummy, &dummy, &dummy, &cs);
	on8(s, c-s);
	width+=cs.width;
	et=GetEUCtype(c);
	s=c;
      }
      else c+=2;
    }
    else if (et==Kana)
    {
      if (Kana!=GetEUCtype(c))
      {
	char buf[BUFSIZE];
	for (int i=0; i<c-s; i+=2)
	  buf[i/2]=0x7f&*(s+i+1);
	XTextExtents(fsKana, buf, (c-s)/2, &dummy, &dummy, &dummy, &cs);
	width+=cs.width;
	et=GetEUCtype(c);
	s=c;
      }
      else c+=2;
    }
    else return width;
  }
}


static int DrawText(int x, int y, char *str)
{
  y+=max(max(fsAscii->max_bounds.ascent,
	     fsKanji->max_bounds.ascent),
	 fsKana->max_bounds.ascent);
  
  int width=0;
  char *c=str;
  char *s=str;
  EUCtype et=GetEUCtype(str);
  
  int dummy;
  XCharStruct cs;
  
  while (1)
  {
    if (et==Ascii)
    {
      if (Ascii!=GetEUCtype(c))
      {
	XSetForeground(d, pxmgc, fore.pixel);
	XSetFont(d, pxmgc, fsAscii->fid);
	XDrawString(d, pxm, pxmgc, width+x, y, s, c-s);
	XTextExtents(fsAscii, s, c-s, &dummy, &dummy, &dummy, &cs);
	width+=cs.width;
	et=GetEUCtype(c);
	s=c;
      }
      else c++;
    }
    else if (et==Esc)
    {
      if (*c=='m')
      {
	c++;
	et=GetEUCtype(c);
	s=c;
      }
      else if (*c=='\0') et=Null;
      else c++;
    }
    else if (et==Kanji)
    {
      if (Kanji!=GetEUCtype(c))
      {
	XSetForeground(d, pxmgc, fore.pixel);
	XSetFont(d, pxmgc, fsKanji->fid);
	off8(s, c-s);
	XDrawString16(d, pxm, pxmgc, width+x, y, (XChar2b*)s, (c-s)/2);
	XTextExtents16(fsKanji, (XChar2b*)s, (c-s)/2, &dummy, &dummy, &dummy, &cs);
	on8(s, c-s);
	width+=cs.width;
	et=GetEUCtype(c);
	s=c;
      }
      else c+=2;
    }
    else if (et==Star)
    {
      if (Star!=GetEUCtype(c))
      {
	XSetForeground(d, pxmgc, star.pixel);
	XSetFont(d, pxmgc, fsKanji->fid);
	off8(s, c-s);
	XDrawString16(d, pxm, pxmgc, width+x, y, (XChar2b*)s, (c-s)/2);
	XTextExtents16(fsKanji, (XChar2b*)s, (c-s)/2, &dummy, &dummy, &dummy, &cs);
	on8(s, c-s);
	width+=cs.width;
	et=GetEUCtype(c);
	s=c;
      }
      else c+=2;
    }
    else if (et==Kana)
    {
      if (Kana!=GetEUCtype(c))
      {
	char buf[BUFSIZE];
	for (int i=0; i<c-s; i+=2)
	  buf[i/2]=*(s+i+1);
	XSetForeground(d, pxmgc, fore.pixel);
	XSetFont(d, pxmgc, fsKana->fid);
	XDrawString(d, pxm, pxmgc, width+x, y, buf, (c-s)/2);
	XTextExtents(fsKana, buf, (c-s)/2, &dummy, &dummy, &dummy, &cs);
	width+=cs.width;
	et=GetEUCtype(c);
	s=c;
      }
      else c+=2;
    }
    else return width;
  }
}


int width=1;
int height=1;


static void calc_window_size(void)
{
  if (app_data.landscape)
  {
    height=GetFontHeight()+12;
    width=0;
    for (int i=0; i<MAX_WHO; i++)
      if (who.who[i])
	width+=GetWidth(who.who[i])+6;
    if (!width) width=1;
  }
  else
  {
    height=(GetFontHeight()+2)*who.N();
    if (!height) height=1;
    width=10;
    int ww;
    for (int i=0; i<MAX_WHO; i++)
      if (who.who[i])
	if (width<(ww=GetWidth(who.who[i])))
	  width=ww+2;
    width+=10;
  }
}


static void resize(void)
{
  if (pxm)   XFreePixmap(d, pxm);
  if (pxmgc) XFreeGC(d, pxmgc);
  calc_window_size();
  pxm=XCreatePixmap(d, w, width, height, DefaultDepth(d, 0));
  pxmgc=XCreateGC(d, pxm, 0, 0);
  
  if (app_data.landscape)
  {
    int x=0;
    for (int i=0; i<MAX_WHO; i++)
      if (who.who[i])
      {
	int ww=GetWidth(who.who[i]);
	XSetForeground(d, pxmgc, ((i%2)?back:back2).pixel);
	XFillRectangle(d, pxm, pxmgc, x, 0, ww+6, height);
	DrawText(x+3, 6, who.who[i]);
	x+=ww+6;
      }
  }
  else
  {
    int y=0;
    for (int i=0; i<MAX_WHO; i++)
      if (who.who[i])
      {
	XSetForeground(d, pxmgc, ((i%2)?back:back2).pixel);
	XFillRectangle(d, pxm, pxmgc, 0, y, width, GetFontHeight()+2);
	DrawText((width-GetWidth(who.who[i]))/2, y+1, who.who[i]);
	y+=GetFontHeight()+2;
      }
  }
}


static void get_window_size(const Window w, unsigned int *width_r, unsigned int *height_r)
{
  Window root;
  int x, y;
  unsigned int border, depth;
  XGetGeometry(d, w, &root, &x, &y, width_r, height_r, &border, &depth);
}


static void redraw(Widget widget, XEvent *e, String *params, Cardinal *num_params)
{
  if (!pxm) resize();
  unsigned int _height, _width;
  get_window_size(w, &_width, &_height);
  if (_width!=width || _height!=height || e==NULL)
  {
    XResizeWindow(d, w, width, height);
    XtVaSetValues
      (shell,
       XtNwidth,     width,
       XtNheight,    height,
       XtNmaxWidth,  width,
       XtNmaxHeight, height,
       XtNminWidth,  width,
       XtNminHeight, height,
       NULL);
    XCopyArea(d, pxm, w, gc, 0, 0, width, height, 0, 0);
  }
  else
    XCopyArea(d, pxm, w, gc,
	      e->xexpose.x,     e->xexpose.y,
	      e->xexpose.width, e->xexpose.height,
	      e->xexpose.x,     e->xexpose.y);
}


static void close_window(Widget widget, XEvent *e, String *params, Cardinal *num_params)
{
  if (e->type==ClientMessage &&
      e->xclient.message_type==atom_wm_protocols &&
      e->xclient.data.l[0]==atom_wm_delete_window)
    exit(0);
}


static void quit(Widget widget, XEvent *e, String *params, Cardinal *num_params)
{
  exit(0);
}


static XtActionsRec actions[]=
{
{"quit",   quit},
{"redraw", redraw},
{"close",  close_window},
};


static String translation=
"<ClientMessage>:   close()\n"
"<ConfigureNotify>: redraw()\n"
"<Key>q:            quit()\n"
"<Key>Escape:       quit()\n"
"<Expose>:          redraw()\n";


SocketStream        soc;
int                 popup_timer=0;


static void read_data(int is_redraw=True)
{
  static f;
  
  int i;
  while ((i=soc.Select(0))>=0)
  {
    if (0<=i && i<MAX_FD)
    {
      char buf[BUFSIZE];
      soc.Gets(soc.SocketFD(), buf, BUFSIZE, 100);
      remove_newline(buf);
      if (strncmp(buf, "## __", 5)==0)
      {
	f=who.N();
	who.Clear();
      }
      else if (strncmp(buf, "## --", 5)==0)
      {
	if (is_redraw)
	{
	  resize();
	  redraw(NULL, NULL, NULL, NULL);
	  if (f<who.N()) XBell(d, 100);
	  XMapWindow(d, w);
	  popup_timer=app_data.popup_time;
	}
      }
      else
      {
	char *c=strchr(buf, '@');
	if (c)
	{
	  *c='\0';
	  who.Add(buf);
	}
      }
    }
    else if (MAX_FD*2<=i && i<MAX_FD*3)
      quit(NULL, NULL, NULL, NULL);
  }
}


static void timer_event(XtPointer client_data, XtIntervalId *id)
{
  if (app_data.popup)
    if (popup_timer>0)
    {
      popup_timer--;
      if (!popup_timer)
	XIconifyWindow(d, w, 0);
    }
  read_data();
  XtAppAddTimeOut(acr, 1000, timer_event, (XtPointer)NULL);
}


int main(int argc, char *argv[])
{
  shell=XtVaAppInitialize
    (&acr, "XItalkBiff", options, XtNumber(options),
     &argc, argv, default_resources, NULL);
  d=XtDisplay(shell);
  XtAppAddActions(acr, actions, XtNumber(actions));
  XtGetApplicationResources(shell, XtPointer(&app_data), resources, XtNumber(resources), NULL, 0);
  XtAugmentTranslations(shell, XtParseTranslationTable(translation));
  
  // socket
  char *server=SocketStream::GetServer(app_data.italkpipe, "localhost");
  US port=SocketStream::GetPort(app_data.italkpipe, 20875);
  if (app_data.help)
  {
    fprintf(stderr,
	    "usage: xitalkbiff [options ...]\n"
	    "       -h, -l, -nl, -i <host[:port]>, -fn <ascii-font>, -fk <kanji-font>,\n"
	    "       -fr <kana-font>, -fg <color>, -bg <color>, -bg2 <color>,\n"
	    "       -p, -pt <popuptime>, -iconic, etc.\n");
    return 1;
  }
  if (soc.Init(SOCKET_STREAM_CLIENT, server, port, True, 100))
    return 1;

   /*
   icon=XCreateBitmapFromData(d, RootWindow(d, 0), (char *)icon_bits, icon_width, icon_height);
   mask=XCreateBitmapFromData(d, RootWindow(d, 0), (char *)mask_bits, mask_width, mask_height);
   XtVaSetValues(shell, XtNiconPixmap, icon, XtNiconMask, mask, NULL);
   */
  
  // font
  if (!(fsAscii=XLoadQueryFont(d, app_data.fontAscii)))
  {
    fprintf(stderr, "Failed to load ascii font.\n");
    return (1);
  }
  if (!(fsKana =XLoadQueryFont(d, app_data.fontKana)))
  {
    fprintf(stderr, "Failed to load kana font.\n");
    fprintf(stderr, "font error\n");
    return (1);
  }
  if (!(fsKanji=XLoadQueryFont(d, app_data.fontKanji)))
  {
    fprintf(stderr, "Failed to load kanji font.\n");
    return (1);
  }

  
  // initial data
  read_data(False);
  calc_window_size();
  XtVaSetValues
    (shell,
     XtNwidth,     width,
     XtNheight,    height,
     XtNmaxWidth,  width,
     XtNmaxHeight, height,
     XtNminWidth,  width,
     XtNminHeight, height,
     XtNinput,     True,
     NULL);
  
  // realize
  XtVaSetValues(shell, XtNinput, True, NULL);
  XtRealizeWidget(shell);
  w=XtWindow(shell);
  atom_wm_protocols    =XInternAtom(d, "WM_PROTOCOLS",     False);
  atom_wm_delete_window=XInternAtom(d, "WM_DELETE_WINDOW", False);
  XSetWMProtocols(d, w, &atom_wm_delete_window, 1);
  gc=XCreateGC(d, w, 0, 0);
  
  // color
  XColor dummy;
  XAllocNamedColor(d, DefaultColormap(d, 0), app_data.foreground,  &fore,  &dummy);
  XSetForeground(d, gc, fore.pixel);
  XAllocNamedColor(d, DefaultColormap(d, 0), app_data.background,  &back,  &dummy);
  XAllocNamedColor(d, DefaultColormap(d, 0), app_data.background2, &back2, &dummy);
  XSetBackground(d, gc, back.pixel);
  XAllocNamedColor(d, DefaultColormap(d, 0), app_data.star,         &star, &dummy);
  
  XtAppAddTimeOut(acr, 1000, timer_event, (XtPointer)NULL);
  
  // signal
  signal(SIGTERM, (void (*)(int))quit);
  signal(SIGINT, (void (*)(int))quit);
  
  // main loop
  XtAppMainLoop(acr);
  return 0;
}
